@@ -13,3 +13,12 @@
 
 #define CRTJMP(pc,sp) __asm__ __volatile__( \
 	"mr 1,%1 ; mtlr %0 ; blr" : : "r"(pc), "r"(sp) : "memory" )
+
+#define GETFUNCSYM(fp, sym, got) __asm__ ( \
+	".hidden " #sym " \n" \
+	"	bl 1f \n" \
+	"	.long " #sym "-. \n" \
+	"1:	mflr %1 \n" \
+	"	lwz %0, 0(%1) \n" \
+	"	add %0, %0, %1 \n" \
+	: "=r"(*(fp)), "=r"((int){0}) : : "memory", "lr" )
