@@ -5,6 +5,11 @@ void __unmapself_sh_nommu(void *, size_t);
 
 #if !defined(__SH3__) && !defined(__SH4__)
 #define __unmapself __unmapself_sh_nommu
+#include "dynlink.h"
+#undef CRTJMP
+#define CRTJMP(pc,sp) __asm__ __volatile__( \
+	"mov.l @%0+,r0 ; mov.l @%0,r12 ; jmp @r0 ; mov %1,r15" \
+	: : "r"(pc), "r"(sp) : "r0", "memory" )
 #include "../../../src/thread/__unmapself.c"
 #undef __unmapself
 extern __attribute__((__visibility__("hidden"))) unsigned __sh_nommu;
