@@ -33,8 +33,7 @@ char *__nl_langinfo_l(nl_item item, locale_t loc)
 	int idx = item & 65535;
 	const char *str;
 
-	if (item == CODESET)
-		return MB_CUR_MAX==1 ? "UTF-8-CODE-UNITS" : "UTF-8";
+	if (item == CODESET) return MB_CUR_MAX==1 ? "ASCII" : "UTF-8";
 	
 	switch (cat) {
 	case LC_NUMERIC:
